/* This file is automatically generated.  DO NOT EDIT! */
extern ld_emulation_xfer_type ld_armelf_linux_eabi_emulation;
extern ld_emulation_xfer_type ld_armelfb_linux_eabi_emulation;

#define EMULATION_LIST \
  &ld_armelf_linux_eabi_emulation, \
  &ld_armelfb_linux_eabi_emulation, \
  0
